(this.webpackJsonp=this.webpackJsonp||[]).push([[4],{1138:function(e,t,n){"use strict";n.r(t);var r,c=n(0),i=n(980),u=n(105),l=n(91);!function(e){e.Added="added",e.Title="title",e.URL="original_url"}(r||(r={}));var a=n(930),o=Object(a.a)((function(){return{sortByOptions:Object(c.useMemo)((function(){return[{value:r.Added,label:"Date Added"},{value:r.Title,label:"Title"},{value:r.URL,label:"URL"}]}),[]),addEntryProps:Object(c.useMemo)((function(){return[{label:"Entry Title",name:"title"},{label:"Entry URL",name:"originalUrl"}]}),[]),api:{list:{useGet:function(){return Object(u.b)("/entry_list")},useAdd:function(){return Object(u.b)("/entry_list",l.a.Post)},useRemove:function(e){return Object(u.b)("/entry_list/".concat(e),l.a.Delete)}},entry:{useGet:function(e,t){return Object(u.b)("/entry_list/".concat(e,"/entries?").concat(t))},useAdd:function(e){return Object(u.b)("/entry_list/".concat(e,"/entries"),l.a.Post)},useRemove:function(e,t){return Object(u.b)("/entry_list/".concat(e,"/entries/").concat(t),l.a.Delete)},useRemoveBulk:function(e){return Object(u.b)("/entry_list/".concat(e,"/entries/batch"),l.a.Delete)}}}}})),s=n(5);t.default=function(){return Object(s.e)(o.Provider,null,Object(s.e)(i.a,{title:"Entry List"}))}}}]);
//# sourceMappingURL=EntryListPlugin.c4b35103ad1660e69c44.js.map